from .preprocessing import (
    asinh_transform,
    auto_logicle_transform,
    scale,
    unscale,
    inverse_transform,
)
from .representation import umap, subcluster
from .result_analysis import count_cell_populations, mean_intensities



from collections import defaultdict
from functools import wraps
from typing import Callable, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from anndata import AnnData
from scipy import stats

from .. import Scyan
from ..utils import _get_subset_indices


def plot_decorator(adata: bool = False):
    def decorator(f: Callable) -> Callable:
        """Decorator that shows a matplotlib figure if the provided 'show' argument is True"""

        @wraps(f)
        def wrapper(model, *args, **kwargs):
            if adata:
                assert isinstance(
                    model, AnnData
                ), f"{f.__name__} first argument has to be anndata.AnnData instance. Received type {type(model)}."
            else:
                assert isinstance(
                    model, Scyan
                ), f"{f.__name__} first argument has to be scyan.Scyan model. Received type {type(model)}."

            res = f(model, *args, **kwargs)
            if kwargs.get("show", True):
                plt.show()
            return res

        return wrapper

    return decorator


def check_population(return_list: bool = False, one: bool = False):
    """Decorator that checks if the provided population (or populations) exists"""

    def decorator(f: Callable) -> Callable:
        @wraps(f)
        def wrapper(
            model: Scyan,
            population: Union[str, List[str]],
            *args,
            obs_key="scyan_pop",
            **kwargs,
        ):
            adata = model if isinstance(model, AnnData) else model.adata
            if adata.obs[obs_key].dtype == "category":
                populations = adata.obs[obs_key].cat.categories.values
            else:
                populations = set(adata.obs[obs_key].values)
            if isinstance(population, str):
                if population not in populations:
                    raise NameError(
                        f"Invalid input population. '{population}' has to be one of {populations}."
                    )
                if return_list:
                    population = [population]
            else:
                if one:
                    raise ValueError(
                        f"Argument 'population' has to be a string. Found {population}."
                    )
                not_found_names = [p for p in population if p not in populations]
                if not_found_names:
                    raise NameError(
                        f"Invalid input population list. {not_found_names} has to be inside {populations}."
                    )
            return f(model, population, *args, obs_key=obs_key, **kwargs)

        return wrapper

    return decorator


def get_palette_others(
    data: pd.DataFrame,
    key: str,
    palette: str = None,
    others: str = "Others",
    value: float = 0.5,
):
    pops = data[key].unique()

    colors = sns.color_palette(palette or "Set1", len(pops))
    colors = dict(zip(pops, colors))
    if others in colors.keys():
        colors[others] = (value, value, value)

    return colors


def ks_statistics(
    adata: AnnData, obs_key: str, populations: List[str], max_obs: int = 5000
):
    statistics = defaultdict(float)

    for pop in populations:
        adata1 = adata[adata.obs[obs_key] == pop]

        if len(populations) > 1:
            other_pops = [p for p in populations if p != pop]
            adata2 = adata[np.isin(adata.obs[obs_key], other_pops)]
        else:
            adata2 = adata[adata.obs[obs_key] != pop]

        adata1 = adata1[_get_subset_indices(adata1.n_obs, max_obs)]
        adata2 = adata2[_get_subset_indices(adata2.n_obs, max_obs)]

        for marker in adata.var_names:
            statistics[marker] += stats.kstest(
                adata1[:, marker].X.flatten(), adata2[:, marker].X.flatten()
            ).statistic

    return statistics


def select_markers(
    adata: AnnData,
    markers: Optional[List[str]],
    n_markers: Optional[int],
    obs_key: str,
    populations: List[str],
    min_markers: int = 2,
):
    MIN_MARKERS_ERROR = f"Provide at least {min_markers} marker(s) to plot or use 'scyan.plot.kde_per_population'"

    if markers is None:
        assert (
            n_markers is not None
        ), "You need to provide a list of markers or a number of markers to be chosen automatically"
        assert n_markers >= min_markers, MIN_MARKERS_ERROR

        statistics = ks_statistics(adata, obs_key, populations)
        statistics = sorted(statistics.items(), key=lambda x: x[1], reverse=True)
        markers = [m for m, _ in statistics[:n_markers]]

    assert len(markers) >= min_markers, MIN_MARKERS_ERROR
    return markers



from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch

from .. import Scyan
from .utils import check_population, plot_decorator


@torch.no_grad()
@plot_decorator()
@check_population(one=True)
def probs_per_marker(
    model: Scyan,
    population: str,
    obs_key: str = "scyan_pop",
    prob_name: str = "Prob",
    vmin_threshold: int = -100,
    figsize: Tuple[float] = (10, 6),
    show: bool = True,
):
    """Interpretability tool: get a group of cells and plot a heatmap of marker probabilities for each population.

    Args:
        model: Scyan model.
        population: Name of one population to interpret. To be valid, the population name has to be in `adata.obs[obs_key]`.
        obs_key: Key to look for population in `adata.obs`. By default, uses the model predictions.
        prob_name: Name to display on the plot.
        vmin_threshold: Minimum threshold for the heatmap colorbar.
        figsize: Pair `(width, height)` indicating the size of the figure.
        show: Whether or not to display the figure.
    """
    u = model(model.adata.obs[obs_key] == population)

    log_probs = model.module.prior.log_prob_per_marker(u)
    mean_log_probs = log_probs.mean(dim=0).cpu().numpy()

    df_probs = pd.DataFrame(
        mean_log_probs,
        columns=model.var_names,
        index=model.pop_names,
    )
    df_probs = df_probs.reindex(
        df_probs.mean().sort_values(ascending=False).index, axis=1
    )
    means = df_probs.mean(axis=1)
    means = means / means.min() * df_probs.values.min()
    df_probs.insert(0, prob_name, means)
    df_probs.insert(1, " ", np.nan)
    df_probs.sort_values(by=prob_name, inplace=True, ascending=False)

    plt.figure(figsize=figsize)
    sns.heatmap(df_probs, cmap="magma", vmin=max(vmin_threshold, mean_log_probs.min()))
    plt.title("Log probabilities per marker for each population")



from collections import defaultdict

import matplotlib.pyplot as plt
import pandas as pd

from .. import Scyan
from .utils import plot_decorator


@plot_decorator()
def pops_hierarchy(model: Scyan, figsize: tuple = (18, 5), show: bool = True) -> None:
    """Plot populations as a tree, where each level corresponds to more detailed populations. To run this function, your knowledge table need to contain at least one population 'level' (see [this tutorial](../../tutorials/advanced/#hierarchical-population-display)), and you need to install `graphviz`.

    Args:
        model: Scyan model.
        figsize: Matplotlib figure size.
        show: Whether or not to display the figure.
    """
    import networkx as nx
    from networkx.drawing.nx_agraph import graphviz_layout

    marker_pop_matrix = model.marker_pop_matrix

    assert isinstance(
        marker_pop_matrix.index, pd.MultiIndex
    ), "To plot population hierarchy, you need a MultiIndex DataFrame. See the documentation for more details."

    G = nx.DiGraph()
    G.add_node(" ")

    def add_nodes(marker_pop_matrix, indices, level, parent=" "):
        if level == -1:
            return

        index = marker_pop_matrix.index.get_level_values(level)
        dict_indices = defaultdict(list)
        for i in indices:
            dict_indices[index[i]].append(i)

        for name, indices in dict_indices.items():
            if not name == parent:
                G.add_node(name)
                G.add_edge(parent, name)
            add_nodes(marker_pop_matrix, indices, level - 1, name)

    add_nodes(
        marker_pop_matrix,
        range(len(marker_pop_matrix)),
        marker_pop_matrix.index.nlevels - 1,
    )

    plt.figure(figsize=figsize)
    pos = graphviz_layout(G, prog="dot")
    nx.draw(G, pos, with_labels=True, arrows=False, node_size=0)



from typing import List, Optional, Union

import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from anndata import AnnData

from .. import Scyan
from ..utils import _subset
from .utils import check_population, get_palette_others, plot_decorator, select_markers


@plot_decorator(adata=True)
@check_population(return_list=True)
def scatter(
    adata: AnnData,
    population: Union[str, List[str]],
    markers: Optional[List[str]] = None,
    n_markers: Optional[int] = 3,
    obs_key: str = "scyan_pop",
    max_obs: int = 2000,
    s: float = 1.0,
    show: bool = True,
) -> None:
    """Display marker expressions on 2D scatter plots with colors per population.
    One scatter plot is displayed for each pair of markers.

    Args:
        adata: An `anndata` object.
        population: One population or a list of population to be colored. To be valid, a population name have to be in `adata.obs[obs_key]`.
        markers: List of markers to plot. If `None`, the list is chosen automatically.
        n_markers: Number of markers to choose automatically if `markers is None`.
        obs_key: Key to look for populations in `adata.obs`. By default, uses the model predictions.
        max_obs: Maximum number of cells per population to be displayed.
        s: Dot marker size.
        show: Whether or not to display the figure.
    """
    markers = select_markers(adata, markers, n_markers, obs_key, population)

    data = adata[:, markers].to_df()
    keys = adata.obs[obs_key].astype(str)
    data["Population"] = np.where(~np.isin(keys, population), "Others", keys)

    pops = list(population) + ["Others"]
    if max_obs is not None:
        groups = data.groupby("Population").groups
        data = data.loc[[i for pop in pops[::-1] for i in _subset(groups[pop], max_obs)]]

    palette = get_palette_others(data, "Population")

    g = sns.PairGrid(data, hue="Population", corner=True, palette=palette, hue_order=pops)
    g.map_offdiag(sns.scatterplot, s=s)
    g.map_diag(sns.histplot)
    g.add_legend()


def umap(
    adata: AnnData,
    color: Union[str, List[str]],
    vmax: Union[str, float] = "p95",
    vmin: Union[str, float] = "p05",
    **scanpy_kwargs: int,
):
    """Plot a UMAP using scanpy.

    !!! note
        If you trained your UMAP with [scyan.tools.umap][] on a subset of cells, it will only display the desired subset of cells.

    Args:
        adata: An `anndata` object.
        color: Marker(s) or `obs` name(s) to color. It can be either just one string, or a list (it will plot one UMAP per element in the list).
        vmax: `scanpy.pl.umap` vmax argument.
        vmin: `scanpy.pl.umap` vmin argument.
        **scanpy_kwargs: Optional kwargs provided to `scanpy.pl.umap`.
    """
    assert isinstance(
        adata, AnnData
    ), f"umap first argument has to be an AnnData object. Received type {type(adata)}."

    assert (
        "X_umap" in adata.obsm_keys()
    ), "Before plotting data, UMAP coordinates need to be computed using 'scyan.tools.umap(...)' (see https://mics-lab.github.io/scyan/api/representation/#scyan.tools.umap)"

    if "has_umap" in adata.obs and not adata.obs.has_umap.all():
        adata = adata[adata.obs.has_umap]

    sc.pl.umap(adata, color=color, vmax=vmax, vmin=vmin, **scanpy_kwargs)


def pop_level(
    model: Scyan,
    group_name: str,
    level_name: str = "level",
    obs_key: str = "scyan_pop",
    **scanpy_kwargs: int,
) -> None:
    """Plot all subpopulations of a group at a certain level on a UMAP (according to the populations levels provided in the knowledge table).

    Args:
        model: Scyan model.
        group_name: The group to look at among the populations of the selected level.
        level_name: Name of the column of the knowledge table containing the names of the grouped populations.
        obs_key: Key of `adata.obs` to access the model predictions.
    """
    adata = model.adata
    mpm = model.marker_pop_matrix

    assert isinstance(
        mpm.index, pd.MultiIndex
    ), "To use this function, you need a MultiIndex DataFrame, see: https://mics-lab.github.io/scyan/tutorials/advanced/#hierarchical-population-display"

    level_names = mpm.index.names[1:]
    assert (
        level_name in level_names
    ), f"Level '{level_name}' unknown. Choose one of: {level_names}"

    base_pops = mpm.index.get_level_values(0)
    group_pops = mpm.index.get_level_values(level_name)
    assert (
        group_name in group_pops
    ), f"Invalid group name '{group_name}'. It has to be one of: {', '.join(group_pops)}."

    valid_populations = [
        pop for pop, group in zip(base_pops, group_pops) if group == group_name
    ]
    key_name = f"{obs_key}_one_level"
    adata.obs[key_name] = pd.Categorical(
        [pop if pop in valid_populations else np.nan for pop in adata.obs[obs_key]]
    )
    umap(
        adata,
        color=key_name,
        title=f"Among {group_name}",
        na_in_legend=False,
        **scanpy_kwargs,
    )



from typing import Callable, List

import torch
from torch import Tensor


def gaussian_kernel(scale: float) -> Callable:
    return lambda dist_squared: torch.exp(-dist_squared / scale)


class LossMMD:
    """Class used to compute the Maximum Mean Discrepancy (MMD) loss for batch-effect correction."""

    def __init__(self, n_markers: int, prior_std: float, mean_na: float):
        """
        Args:
            n_markers: Number of markers in the table.
            prior_std: Model standard deviation $\sigma$ for $H$.
            mean_na: Mean number of NA per row in the table.
        """
        self.scales = self.get_heuristic_scales(n_markers, prior_std, mean_na)
        self.kernels = [gaussian_kernel(scale) for scale in self.scales]

    def get_heuristic_scales(
        self, n_markers: int, prior_std: float, mean_na: float
    ) -> List[float]:
        internal_scale = 2 * n_markers * prior_std**2
        return [
            0.25 * internal_scale,
            internal_scale,
            4 * internal_scale,
        ]

    def one_kernel_mmd(
        self, kernel: Callable, dxx: Tensor, dxy: Tensor, dyy: Tensor
    ) -> Tensor:
        return (kernel(dxx) - 2 * kernel(dxy) + kernel(dyy)).mean()

    def __call__(self, x: Tensor, y: Tensor) -> Tensor:
        """Compute the MMD loss.

        Args:
            x: Tensor of size $(B, M)$.
            y: Tensor of size $(B, M)$.

        Returns:
            Loss term as a one-element Tensor.
        """
        xx, yy, xy = x.mm(x.T), y.mm(y.T), x.mm(y.T)

        x2 = (x**2).sum(dim=-1)
        y2 = (y**2).sum(dim=-1)

        dxx = x2[:, None] + x2[None, :] - 2 * xx
        dxy = x2[:, None] + y2[None, :] - 2 * xy
        dyy = y2[:, None] + y2[None, :] - 2 * yy

        return sum(self.one_kernel_mmd(kernel, dxx, dxy, dyy) for kernel in self.kernels)



from .density import kde_per_population
from .heatmap import probs_per_marker
from .dot import scatter, umap, pop_level
from .graph import pops_hierarchy
from .expressions import pop_expressions, pops_expressions, boxplot_expressions



import logging
import warnings
from functools import wraps
from pathlib import Path
from typing import Callable, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from anndata import AnnData
from pandas.api.types import is_numeric_dtype
from torch import Tensor

log = logging.getLogger(__name__)

warnings.filterwarnings("ignore", message="Transforming to str index")
warnings.filterwarnings("ignore", message=r".*Trying to modify attribute[\s\S]*")
warnings.filterwarnings("ignore", message=r".*No data for colormapping provided[\s\S]*")


def _root_path() -> Path:
    """Get the library root path

    Returns:
        `scyan` library root path
    """
    return Path(__file__).parent.parent


def _wandb_plt_image(fun: Callable, figsize: Tuple[int, int] = [7, 5]):
    """Transform a matplotlib figure into a wandb Image

    Args:
        fun: Function that makes the plot - do not plt.show().
        figsize: Matplotlib figure size.

    Returns:
        The wandb Image to be logged.
    """

    import io

    import wandb
    from PIL import Image

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["figure.autolayout"] = True
    plt.figure()
    fun()
    img_buf = io.BytesIO()
    plt.savefig(img_buf, format="png")
    return wandb.Image(Image.open(img_buf))


def _subset(indices: List[str], max_obs: int):
    if len(indices) < max_obs:
        return indices
    return indices[np.random.choice(len(indices), max_obs, replace=False)]


def _markers_to_indices(model, markers: List[str]) -> Tensor:
    return torch.tensor(
        [model.var_names.get_loc(marker) for marker in markers],
        dtype=int,
    )


def _pop_to_index(model, pop: str):
    assert pop in model.pop_names, f"Found invalid population name '{pop}'"
    return model.pop_names.get_loc(pop)


def _pops_to_indices(model, pops: List[str]) -> Tensor:
    return torch.tensor([_pop_to_index(model, pop) for pop in pops], dtype=int)


def _get_subset_indices(n_obs, n_cells: Union[int, None]):
    if n_cells is None or n_cells >= n_obs:
        return np.arange(n_obs)
    return np.random.choice(n_obs, size=n_cells, replace=False)


def _process_pop_sample(model, pop: Union[str, List[str], int, Tensor, None] = None):
    if isinstance(pop, str):
        return _pop_to_index(model, pop)
    if isinstance(pop, list):
        return _pops_to_indices(model, pop)
    else:
        return pop


def _requires_fit(f: Callable) -> Callable:
    """Make sure the model has been trained"""

    @wraps(f)
    def wrapper(model, *args, **kwargs):
        assert (
            model._is_fitted
        ), "The model has to be trained first, consider running 'model.fit()'"
        return f(model, *args, **kwargs)

    return wrapper


def _add_level_predictions(model, obs_key: str) -> None:
    mpm: pd.DataFrame = model.marker_pop_matrix
    adata: AnnData = model.adata

    level_names = mpm.index.names[1:]
    obs_keys = [f"{obs_key}_{name}" for name in level_names]

    for i, new_obs_key in enumerate(obs_keys):
        pop_dict = {pop: levels_pops[i] for pop, *levels_pops in mpm.index}
        adata.obs[new_obs_key] = adata.obs[obs_key].map(pop_dict).astype("category")


def _get_pop_index(pop: str, marker_pop_matrix: pd.DataFrame):
    for i in range(marker_pop_matrix.index.nlevels - 1, -1, -1):
        if pop in marker_pop_matrix.index.get_level_values(i):
            return i
    raise BaseException(f"Population {pop} not found.")


def _check_is_processed(X: np.ndarray) -> None:
    assert (
        np.abs(X).max() < 1e3
    ), "The provided values are very high: have you run preprocessing first? E.g., consider running 'scyan.tools.asinh_transform' or 'scyan.tools.auto_logicle_transform' (see our tutorial: https://mics-lab.github.io/scyan/tutorials/preprocessing/)"


def _validate_inputs(adata: AnnData, df: pd.DataFrame):
    assert isinstance(
        adata, AnnData
    ), f"The provided adata has to be an AnnData object (https://anndata.readthedocs.io/en/latest/), found {type(adata)}."

    assert isinstance(
        df, pd.DataFrame
    ), f"The marker-population matrix has to be a pandas DataFrame, found {type(df)}"

    not_found_columns = [c for c in df.columns if c not in adata.var_names]

    assert (
        not not_found_columns
    ), f"All column names from the marker-population table have to be a known marker from adata.var_names. Missing {not_found_columns}."

    if not df.dtypes.apply(is_numeric_dtype).all():
        log.warn(
            "Some columns of the marker-population table are not numeric / NaN. Every non-numeric value will be transformed into NaN."
        )
        df = df.apply(pd.to_numeric, errors="coerce")

    ratio_nan = df.isna().values.mean()
    if ratio_nan > 0.7:
        log.warn(
            f"Found {ratio_nan:.1%} of NA in the table, which is very high. If this is intended, just ignore the warning."
        )

    X = adata[:, df.columns].X

    _check_is_processed(X)

    if np.abs(X.std(axis=0) - 1).max() > 0.2:
        log.warn(
            "It seems that the data is not standardised. We advise using scaling (scyan.tools.scale) before initializing the model."
        )

    duplicates = df.duplicated()
    if duplicates.any():
        duplicates_names = duplicates[duplicates].index.get_level_values(0)
        log.warn(
            f"Found duplicate populations in the knowledge matrix. We advise updating or removing the following rows: {', '.join(map(str, duplicates_names))}"
        )

    if isinstance(df.index, pd.MultiIndex):
        assert (
            not df.index.to_frame().isna().any().any()
        ), "One or multiple population name(s) are NaN, you have to name them."

    ratio_non_standard = 1 - ((df**2 == 1) | df.isna()).values.mean()
    if ratio_non_standard > 0.15:
        log.warn(
            f"Found a significant proportion ({ratio_non_standard:.1%}) of non-standard values in the knowledge table. Scyan expects to find mostly -1/1/NA in the table, even though any other numerical value is accepted. If this is intended, just ignore the warning, else correct the table using mainly -1, 1 and NA (to denote negative expressions, positive expressions, or not-applicable respectively)."
        )

    return adata, df



import logging
from typing import Dict, List, Optional, Union

import fcsparser
import fcswrite
import numpy as np
import pandas as pd
from anndata import AnnData
from pandas.api.types import is_numeric_dtype

log = logging.getLogger(__name__)


def read_fcs(
    path: str, obs_names: Optional[List[str]] = None, channel_suffix: Optional[str] = "S"
) -> AnnData:
    """Read a FCS file and return an AnnData instance.

    Args:
        path: Path to the FCS file that has to be read.
        obs_names: Optional list of channel names that has to be considered as an observation (i.e., inside `adata.obs`). By default, choose it automatically.
        channel_suffix: Suffix for the channel naming convention, i.e. `"S"` for "PnS", or `"N"` for "PnN". If `None`, keep the raw names.

    Returns:
        `AnnData` instance containing the FCS data.
    """
    meta, data = fcsparser.parse(path)

    if channel_suffix is not None:
        data.columns = [
            meta.get(f"$P{i + 1}{channel_suffix}", c) for i, c in enumerate(data.columns)
        ]

    if obs_names is None:
        obs_names = [
            c
            for i, c in enumerate(data.columns)
            if not f"$P{i + 1}{channel_suffix}" in meta
        ]
    elif not all(c in data.columns for c in obs_names):
        log.warn(
            f"The following observations were not found: {', '.join([c for c in obs_names if not c in data.columns])}"
        )
        obs_names = [c for c in obs_names if c in data.columns]

    var_names = [c for c in data.columns if not c in obs_names]

    return AnnData(
        X=data[var_names].values,
        var=pd.DataFrame(index=var_names),
        obs=data[obs_names],
        dtype=np.float32,
    )


def _to_df(adata: AnnData, layer: Optional[str] = None) -> pd.DataFrame:
    df = pd.concat([adata.to_df(layer), adata.obs], axis=1)

    for key in adata.obsm:
        names = [f"{key}{i+1}" for i in range(adata.obsm[key].shape[1])]
        df[names] = adata.obsm[key]

    return df


def write_fcs(
    adata: AnnData,
    path: str,
    layer: Optional[str] = None,
    columns_to_numeric: Optional[List] = None,
    **fcswrite_kwargs: int,
) -> Union[None, Dict]:
    """Based on a `AnnData` object, it writes a FCS file that contains (i) all the markers intensities, (ii) every numeric column of `adata.obs`, and (iii) all `adata.obsm` variables.

    !!! note
        As the FCS format doesn't support strings, some observations will not be kept in the FCS file.

    Args:
        adata: `AnnData` object to save.
        path: Path to write the file.
        layer: Name of the `adata` layer from which intensities will be extracted. If `None`, uses `adata.X`.
        columns_to_numeric: List of **non-numerical** column names from `adata.obs` that should be kept, by transforming them into integers. Note that you don't need to list the numerical columns, that are written inside the FCS by default.
        **fcswrite_kwargs: Optional kwargs provided to `fcswrite.write_fcs`.

    Returns:
        If `columns_to_numeric` is `None`, returns nothing. Else, return a dict whose keys are the observation column names being transformed, and the values are ordered lists of the label encoded classes. E.g., `{"batch": ["b1", "b2"]}` means that the batch `"b1"` was encoded by 0, and `"b2"` by 1.
    """
    df = _to_df(adata, layer)
    dict_classes = {}
    columns_removed = []

    for column in df.columns:
        if is_numeric_dtype(df[column].dtype):
            continue
        try:
            df[column] = pd.to_numeric(df[column].values)
            continue
        except:
            from sklearn.preprocessing import LabelEncoder

            if columns_to_numeric is not None and column in columns_to_numeric:
                le = LabelEncoder()
                df[column] = le.fit_transform(df[column].values)
                dict_classes[column] = list(le.classes_)
            else:
                del df[column]
                columns_removed.append(column)

    log.info(f"Found {len(df.columns)} features: {', '.join(df.columns)}.")
    if columns_removed:
        log.info(
            f"FCS does not support strings, so the following columns where removed: {', '.join(columns_removed)}.\nIf you want to keep these str observations, use the 'columns_to_numeric' argument to encod them."
        )

    fcswrite.write_fcs(str(path), list(df.columns), df.values, **fcswrite_kwargs)

    if columns_to_numeric is not None:
        return dict_classes


def write_csv(
    adata: AnnData,
    path: str,
    layer: Optional[str] = None,
) -> Union[None, Dict]:
    """Based on a `AnnData` object, it writes a CSV file that contains (i) all the markers intensities, (ii) every numeric column of `adata.obs`, and (iii) all `adata.obsm` variables.

    Args:
        adata: `AnnData` object to save.
        path: Path to write the file.
        layer: Name of the `adata` layer from which intensities will be extracted. If `None`, uses `adata.X`.
    """
    _to_df(adata, layer).to_csv(path)



import importlib.metadata
import logging

import pytorch_lightning as pl
import scanpy as sc

from .model import Scyan
from . import data, plot, tools, utils
from .io import read_fcs, write_fcs, write_csv

__version__ = importlib.metadata.version("scyan")

logging.basicConfig(level=logging.INFO)
sc.set_figure_params(facecolor="white", fontsize=10)
pl.seed_everything(0)



from typing import List, Optional, Union

import numpy as np
import pandas as pd
import seaborn as sns
from anndata import AnnData

from ..utils import _get_subset_indices
from .utils import check_population, get_palette_others, plot_decorator, select_markers


@plot_decorator(adata=True)
@check_population(return_list=True)
def kde_per_population(
    adata: AnnData,
    population: Union[str, List[str]],
    markers: Optional[List[str]] = None,
    obs_key: str = "scyan_pop",
    n_markers: Optional[int] = 3,
    n_cells: Optional[int] = 100000,
    ncols: int = 2,
    var_name: str = "Marker",
    value_name: str = "Expression",
    show: bool = True,
):
    """Plot Kernel-Density-Estimation for each provided population and for multiple markers.

    Args:
        adata: An `anndata` object.
        population: One population or a list of population to interpret. To be valid, a population name have to be in `adata.obs[obs_key]`.
        markers: List of markers to plot. If `None`, the list is chosen automatically.
        obs_key: Key to look for populations in `adata.obs`. By default, uses the model predictions.
        n_markers: Number of markers to choose automatically if `markers is None`.
        n_cells: Number of cells to be considered for the heatmap (to accelerate it when $N$ is very high). If `None`, consider all cells.
        ncols: Number of figures per row.
        var_name: Name displayed on the graphs.
        value_name: Name displayed on the graphs.
        show: Whether or not to display the figure.
    """
    indices = _get_subset_indices(adata.n_obs, n_cells)
    adata = adata[indices]

    markers = select_markers(adata, markers, n_markers, obs_key, population, 1)

    df = adata.to_df()

    keys = adata.obs[obs_key]
    df[obs_key] = np.where(~np.isin(keys, population), "Others", keys)

    df = pd.melt(
        df,
        id_vars=[obs_key],
        value_vars=markers,
        var_name=var_name,
        value_name=value_name,
    )

    sns.displot(
        df,
        x=value_name,
        col=var_name,
        hue=obs_key,
        col_wrap=ncols,
        kind="kde",
        common_norm=False,
        facet_kws=dict(sharey=False),
        palette=get_palette_others(df, obs_key),
        hue_order=sorted(df[obs_key].unique(), key="Others".__eq__),
    )



from typing import List, Optional, Union

import pandas as pd
from anndata import AnnData


def count_cell_populations(
    adata: AnnData, groupby: str, normalize: bool = False, obs_key: str = "scyan_pop"
) -> pd.DataFrame:
    """Count for each patient (or group) the number of cells for each population.

    Args:
        adata: An `AnnData` object.
        groupby: Key of `adata.obs` used to create groups (e.g. the patient ID).
        normalize: If `True`, returns percentage instead of counts.
        obs_key: Key of `adata.obs` containing the population names (or the values to count).

    Returns:
        A DataFrame of counts (one row per group, one column per population).
    """
    grouped = adata.obs.groupby(groupby)
    counts = grouped[obs_key].apply(lambda s: s.value_counts(normalize))
    return counts.unstack(level=-1)


def mean_intensities(
    adata: AnnData,
    groupby: Union[str, List[str]] = "scyan_pop",
    unstack_join: Optional[str] = None,
    layer: Optional[str] = None,
    obsm: Optional[str] = None,
    obsm_names: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Compute the Mean Metal Intensity (MMI) or Mean Fluorescence Intensity (MFI) per population.

    Args:
        adata: An `AnnData` object.
        groupby: Key(s) of `adata.obs` used to create groups. E.g., `scyan_pop` creates group of populations. It can be one string, or a list of strings: for instance, use `["id", "scyan_pop"]` computes MMI per population for each ID. If you want to get one row per `id`, use `unstack_join`.
        unstack_join: If not `None`, flattens the biomarkers into one series (or one row per group if `groupby` is a list) and uses `unstack_join` to join the names of the multi-level columns. For instance, `' '` can be a good choice.
        layer: In which `adata.layers` we get fluorescence intensities. By default, it uses `adata.X`.
        obsm: In which `adata.obsm` we get fluorescence intensities. By default, it uses `adata.X`. If not `None` then `obsm_names` is required too.
        obsm_names: Ordered list of names in `adata.obsm[obsm]` if `obsm` was provided.

    Returns:
        A DataFrame of MFI. If `groupby` was a list, it is a multi-index dataframe.
    """
    if isinstance(groupby, str):
        groupby = [groupby]

    if obsm is not None:
        assert (
            layer is None
        ), "You must choose between 'obsm' and 'layer', do not use both."

        df = pd.DataFrame(data=adata.obsm[obsm], columns=obsm_names)
    else:
        df = adata.to_df(layer)

    for group in groupby:
        df[group] = adata.obs[group].values

    res = df.groupby(groupby).mean()

    if unstack_join is None:
        return res

    res = res.unstack(level=-1)
    if isinstance(res, pd.Series):
        res.index = [unstack_join.join(row).strip() for row in res.index.values]
    else:
        res.columns = [unstack_join.join(col).strip() for col in res.columns.values]
    return res



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scyan', 'scyan.data', 'scyan.module', 'scyan.plot', 'scyan.tools']

package_data = \
{'': ['*']}

install_requires = \
['FlowUtils>=1.0.0,<2.0.0',
 'fcsparser>=0.2.4,<0.3.0',
 'fcswrite>=0.6.2,<0.7.0',
 'llvmlite>=0.38.1,<0.39.0',
 'matplotlib>=3.5.2,<4.0.0',
 'pytorch-lightning>=1.6.4,<2.0.0',
 'scanpy>=1.9.1,<2.0.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'umap-learn>=0.5.3,<0.6.0']

extras_require = \
{'dev': ['wandb>=0.12.20,<0.13.0',
         'hydra-core>=1.2.0,<2.0.0',
         'hydra-colorlog>=1.2.0,<2.0.0',
         'hydra-optuna-sweeper>=1.2.0,<2.0.0',
         'pytest>=7.1.2,<8.0.0',
         'ipykernel>=6.15.0,<7.0.0',
         'ipywidgets>=7.7.1,<8.0.0',
         'isort>=5.10.1,<6.0.0',
         'black>=22.6.0,<23.0.0',
         'imblearn>=0.0,<0.1'],
 'discovery': ['leidenalg>=0.8.10,<0.9.0'],
 'docs': ['mkdocs-material>=8.3.9,<9.0.0',
          'mkdocstrings>=0.19.0,<0.20.0',
          'mkdocstrings-python>=0.7.1,<0.8.0',
          'mkdocs-jupyter>=0.21.0,<0.22.0']}

setup_kwargs = {
    'name': 'scyan',
    'version': '1.2.2',
    'description': 'Single-cell Cytometry Annotation Network',
    'long_description': '<p align="center">\n  <img src="https://github.com/MICS-Lab/scyan/raw/master/docs/assets/logo.png" alt="scyan_logo" width="500"/>\n</p>\n\n[![PyPI](https://img.shields.io/pypi/v/scyan.svg)](https://pypi.org/project/scyan)\n[![Docs](https://img.shields.io/badge/docs-mkdocs-blue)](https://mics-lab.github.io/scyan/)\n![Build](https://github.com/MICS-Lab/scyan/workflows/ci/badge.svg)\n[![Code Style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/python/black)\n[![Downloads](https://pepy.tech/badge/scyan)](https://pepy.tech/project/scyan)\n[![License](https://img.shields.io/pypi/l/scyan.svg)](https://github.com/MICS-Lab/scyan/blob/master/LICENSE)\n[![Imports: isort](https://img.shields.io/badge/imports-isort-blueviolet)](https://pycqa.github.io/isort/)\n\nScyan stands for **S**ingle-cell **Cy**tometry **A**nnotation **N**etwork. Based on biological knowledge prior, it provides a fast cell population annotation without requiring any training label. Scyan is an interpretable model that also corrects batch-effect and can be used for debarcoding, cell sampling, and population discovery.\n\n# Documentation\n\nThe [complete documentation can be found here](https://mics-lab.github.io/scyan/). It contains installation guidelines, tutorials, a description of the API, etc.\n\n# Overview\n\nScyan is a Bayesian probabilistic model composed of a deep invertible neural network called a normalizing flow (the function $f_{\\phi}$). It maps a latent distribution of cell expressions into the empirical distribution of cell expressions. This cell distribution is a mixture of gaussian-like distributions representing the sum of a cell-specific and a population-specific term. Also, interpretability and batch effect correction are based on the model latent space — more details in the article\'s Methods section.\n\n<p align="center">\n  <img src="https://github.com/MICS-Lab/scyan/raw/master/docs/assets/overview.png" alt="overview_image"/>\n</p>\n\n# Getting started\n\nScyan can be installed on every OS with `pip` or [`poetry`](https://python-poetry.org/docs/).\n\nOn macOS / Linux, `python>=3.8,<3.11` is required, while `python>=3.8,<3.10` is required on Windows. The preferred Python version is `3.9`.\n\n## Install from PyPI (recommended)\n\n```bash\npip install scyan\n```\n\n## Install locally (if you want to contribute)\n\n> Advice (optional): We advise creating a new environment via a package manager (except if you use Poetry, which will automatically create the environment). For instance, you can create a new `conda` environment:\n>\n> ```bash\n> conda create --name scyan python=3.9\n> conda activate scyan\n> ```\n\nClone the repository and move to its root:\n\n```bash\ngit clone https://github.com/MICS-Lab/scyan.git\ncd scyan\n```\n\nChoose one of the following, depending on your needs (it should take at most a few minutes):\n\n```bash\npip install .                           # pip minimal installation (library only)\npip install -e \'.[dev,docs,discovery]\'  # pip installation in editable mode\npoetry install -E \'dev docs discovery\'  # poetry installation in editable mode\n```\n\n## Basic usage / Demo\n\n```py\nimport scyan\n\nadata, marker_pop_matrix = scyan.data.load("aml")\n\nmodel = scyan.Scyan(adata, marker_pop_matrix)\nmodel.fit()\nmodel.predict()\n```\n\nThis code should run in approximately 40 seconds (once the dataset is loaded).\nFor more usage demo, read the [tutorials](https://mics-lab.github.io/scyan/tutorials/usage/) or the complete [documentation](https://mics-lab.github.io/scyan/).\n\n# Technical description\n\nScyan is a **Python** library based on:\n\n- _Pytorch_, a deep learning framework\n- _AnnData_, a data library that works nicely with single-cell data\n- _Pytorch Lighning_, for model training\n- _Hydra_, for project configuration (optional)\n- _Weight & Biases_, for model monitoring (optional)\n\n# Project layout\n\n    .github/      # Github CI and templates\n    config/       # Hydra configuration folder (optional use)\n    data/         # Data folder containing adata files and csv tables\n    docs/         # The folder used to build the documentation\n    scripts/      # Scripts to reproduce the results from the article\n    tests/        # Folder containing tests\n    scyan/                    # Library source code\n        data/                 # Folder with data-related functions and classes\n            datasets.py       # Load and save datasets\n            tensors.py        # Pytorch data-related classes for training\n        module/               # Folder containing neural network modules\n            coupling_layer.py # Coupling layer\n            distribution.py   # Prior distribution (called U in the article)\n            mmd.py            # Maximum Mean Discrepancy implementation\n            real_nvp.py       # Normalizing Flow\n            scyan_module      # Core module\n        plot/                 # Plotting tools\n            ...\n        tools/\n            ...               # Preprocessing tools and more\n        model.py              # Scyan model class\n        utils.py              # Misc functions\n    .gitattributes\n    .gitignore\n    CONTRIBUTING.md   # To read before contributing\n    LICENSE\n    mkdocs.yml        # The docs configuration file\n    poetry.lock\n    pyproject.toml    # Dependencies, project metadata, and more\n    README.md\n    setup.py          # Setup file, see `pyproject.toml`\n\n# Cite us\n\nOur paper is not published yet. Meanwhile, you can [read our preprint on arXiv](https://arxiv.org/abs/2208.05745).\n',
    'author': 'Blampey Quentin',
    'author_email': 'quentin.blampey@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://mics-lab.github.io/scyan/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import logging
from typing import List, Optional, Tuple, Union

import pytorch_lightning as pl
import torch
from torch import Tensor, distributions, nn

from . import LossMMD, PriorDistribution, RealNVP

log = logging.getLogger(__name__)


class ScyanModule(pl.LightningModule):
    """Core logic contained inside the main class [Scyan][scyan.Scyan]. Do not use this class directly.

    Attributes:
        real_nvp (RealNVP): The Normalizing Flow (a [RealNVP][scyan.module.RealNVP] object)
        prior (PriorDistribution): The prior $U$ (a [PriorDistribution][scyan.module.PriorDistribution] object)
        loss_mmd (LossMMD): The MMD loss (a [LossMMD][scyan.module.LossMMD] object)
        pi_logit (Tensor): Logits used to learn the population weights
    """

    pi_logit_ratio: float = 100  # To learn pi logits faster

    def __init__(
        self,
        rho: Tensor,
        n_covariates: int,
        other_batches: List[int],
        hidden_size: int,
        n_hidden_layers: int,
        n_layers: int,
        prior_std: float,
        temperature: float,
        mmd_max_samples: int,
        batch_ref_id: Optional[int],
    ):
        """
        Args:
            rho: Tensor $\rho$ representing the knowledge table.
            n_covariates: Number of covariates $M_c$ considered.
            other_batches: List of batches that are not the reference.
            hidden_size: MLP (`s` and `t`) hidden size.
            n_hidden_layers: Number of hidden layers for the MLP (`s` and `t`).
            n_layers: Number of coupling layers.
            prior_std: Standard deviation $\sigma$ of the cell-specific random variable $H$.
            temperature: Temperature to favour small populations.
            mmd_max_samples: Maximum number of samples to give to the MMD.
            batch_ref_id: ID corresponding to the reference batch.
        """
        super().__init__()
        self.save_hyperparameters(ignore=["rho", "n_covariates", "other_batches"])

        self.n_pops, self.n_markers = rho.shape
        self.register_buffer("rho", rho)

        self.rho_mask = self.rho.isnan()
        self.rho[self.rho_mask] = 0
        self.mean_na = self.rho_mask.sum() / self.n_markers

        self.pi_logit = nn.Parameter(torch.zeros(self.n_pops))

        self.real_nvp = RealNVP(
            self.n_markers + n_covariates,
            self.hparams.hidden_size,
            self.n_markers,
            self.hparams.n_hidden_layers,
            self.hparams.n_layers,
        )

        self.prior = PriorDistribution(
            self.rho, self.rho_mask, self.hparams.prior_std, self.n_markers
        )

        self.other_batches = other_batches
        self.loss_mmd = LossMMD(self.n_markers, self.hparams.prior_std, self.mean_na)

    def forward(self, x: Tensor, covariates: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """Forward implementation, going through the complete flow $f_{\phi}$.

        Args:
            x: Inputs of size $(B, M)$.
            covariates: Covariates of size $(B, M_c)$

        Returns:
            Tuple of (outputs, covariates, lod_det_jacobian sum)
        """
        return self.real_nvp(x, covariates)

    @torch.no_grad()
    def inverse(self, u: Tensor, covariates: Tensor) -> Tensor:
        """Go through the flow in reverse direction, i.e. $f_{\phi}^{-1}$.

        Args:
            u: Latent expressions of size $(B, M)$.
            covariates: Covariates of size $(B, M_c)$

        Returns:
            Outputs of size $(B, M)$.
        """
        return self.real_nvp.inverse(u, covariates)

    @property
    def prior_z(self) -> distributions.Distribution:
        """Population prior, i.e. $Categorical(\pi)$.

        Returns:
            Distribution of the population index.
        """
        return distributions.Categorical(self.pi)

    @property
    def log_pi(self) -> Tensor:
        """Log population weights $log \; \pi$."""
        return torch.log_softmax(self.pi_logit_ratio * self.pi_logit, dim=0)

    @property
    def pi(self) -> Tensor:
        """Population weights $\pi$"""
        return torch.exp(self.log_pi)

    def log_pi_temperature(self, T: float) -> Tensor:
        """Compute the log weights with temperature $log \; \pi^{(-T)}$

        Args:
            T: Temperature.

        Returns:
            Log weights with temperature.
        """
        return torch.log_softmax(self.pi_logit_ratio * self.pi_logit / T, dim=0).detach()

    @torch.no_grad()
    def sample(
        self,
        n_samples: int,
        covariates: Tensor,
        z: Union[int, Tensor, None] = None,
        return_z: bool = False,
    ) -> Tuple[Tensor, Tensor]:
        """Sampling cell-marker expressions.

        Args:
            n_samples: Number of cells to sample.
            covariates: Tensor of covariates.
            z: Either one population index or a Tensor of population indices. If None, sampling from all populations.
            return_z: Whether to return the population Tensor.

        Returns:
            Sampled cells expressions and, if `return_z`, the populations associated to these cells.
        """
        if z is None:
            z = self.prior_z.sample((n_samples,))
        elif isinstance(z, int):
            z = torch.full((n_samples,), z)
        elif isinstance(z, torch.Tensor):
            z = z.to(int)
        else:
            raise ValueError(
                f"z has to be 'None', an 'int' or a 'torch.Tensor'. Found type {type(z)}."
            )

        u = self.prior.sample(z)
        x = self.inverse(u, covariates)

        return (x, z) if return_z else x

    def compute_probabilities(
        self, x: Tensor, covariates: Tensor, use_temp: bool = False
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Compute probabilities used in the loss function.

        Args:
            x: Inputs of size $(B, M)$.
            covariates: Covariates of size $(B, M_c)$.

        Returns:
            Log probabilities of size $(B, P)$, the log det jacobian and the latent expressions of size $(B, M)$.
        """
        u, _, ldj_sum = self(x, covariates)

        log_pi = (
            self.log_pi_temperature(-self.hparams.temperature)
            if use_temp
            else self.log_pi
        )

        log_probs = self.prior.log_prob(u) + log_pi  # size N x P

        return log_probs, ldj_sum, u

    def batch_correction_mmd(self, u1, pop1, u2, pop2):
        n_samples = min(len(u1), len(u2), self.hparams.mmd_max_samples)

        if n_samples < 500:
            log.warn(f"Correcting batch effect with few samples ({n_samples})")

        pop_weights = 1 / self.pi.detach()

        indices1 = torch.multinomial(pop_weights[pop1], n_samples)
        indices2 = torch.multinomial(pop_weights[pop2], n_samples)

        return self.loss_mmd(u1[indices1], u2[indices2])

    def get_mmd_inputs(self, u: Tensor, batches: Tensor, probs: Tensor, b: int):
        condition = batches == b
        return u[condition], probs[condition].argmax(dim=1)

    def losses(
        self,
        x: Tensor,
        covariates: Tensor,
        batches: Tensor,
        use_temp: bool,
    ) -> Tuple[Tensor, Tensor]:
        """Compute the module loss for one mini-batch.

        Args:
            x: Inputs of size $(B, M)$.
            covariates: Covariates of size $(B, M_c)$.
            batches: Batch information used to correct batch-effect (tensor of size $(B)$)
            use_temp: Whether to consider temperature is the KL term.

        Returns:
            The KL loss term and the MMD loss term.
        """
        log_probs, ldj_sum, u = self.compute_probabilities(x, covariates, use_temp)

        kl = -(torch.logsumexp(log_probs, dim=1) + ldj_sum).mean()

        if self.hparams.batch_ref_id is None:
            return kl, 0

        u_ref, pop_ref = self.get_mmd_inputs(
            u, batches, log_probs, self.hparams.batch_ref_id
        )

        mmd = sum(
            self.batch_correction_mmd(
                u_ref, pop_ref, *self.get_mmd_inputs(u, batches, log_probs, other)
            )
            for other in self.other_batches
        )

        return kl, mmd



from typing import List, Optional, Sized, Tuple, Union

import numpy as np
import pandas as pd
import torch
from anndata import AnnData
from torch import Tensor


class AdataDataset(torch.utils.data.Dataset):
    """Pyorch Dataset"""

    def __init__(self, x: Tensor, covariates: Tensor, batch: Tensor):
        super().__init__()
        self.x = x
        self.covariates = covariates
        self.batch = batch

    def __getitem__(self, index):
        return self.x[index], self.covariates[index], self.batch[index]

    def __len__(self):
        return len(self.x)


class RandomSampler(torch.utils.data.Sampler):
    """Random sampling during training. It stops the epoch when we reached `max_samples` samples (if provided)."""

    def __init__(self, data_source: Sized, max_samples: Optional[int]):
        self.data_source = data_source
        self.n_samples = len(data_source)
        self.max_samples = self.n_samples if max_samples is None else max_samples

    def __iter__(self):
        return iter(torch.randperm(self.n_samples)[: self.max_samples])

    def __len__(self):
        return self.max_samples


def _prepare_data(
    adata: AnnData,
    markers: List[str],
    batch_key: Union[str, int, None],
    batch_ref: Union[str, int, None],
    categorical_covariate_keys: List[str],
    continuous_covariate_keys: List[str],
) -> Tuple[Tensor, Tensor, Tensor]:
    """Initialize the data and the covariates"""
    x = torch.tensor(adata[:, markers].X)

    if (batch_key is not None) and batch_key not in categorical_covariate_keys:
        categorical_covariate_keys.append(batch_key)

    for key in categorical_covariate_keys:  # enforce dtype category
        adata.obs[key] = adata.obs[key].astype("category")

    categorical_covariate_embedding = (
        pd.get_dummies(adata.obs[categorical_covariate_keys]).values
        if categorical_covariate_keys
        else np.empty((adata.n_obs, 0))
    )

    continuous_covariate_embedding = (
        adata.obs[continuous_covariate_keys].values
        if continuous_covariate_keys
        else np.empty((adata.n_obs, 0))
    )

    covariates = np.concatenate(
        [
            categorical_covariate_embedding,
            continuous_covariate_embedding,
        ],
        axis=1,
    )
    covariates = torch.tensor(
        covariates,
        dtype=torch.float32,
    )

    if batch_key is not None:
        batch_to_id = {b: i for i, b in enumerate(adata.obs[batch_key].cat.categories)}
        batches = torch.tensor([batch_to_id[b] for b in adata.obs[batch_key]])
        other_batches = [
            batch_to_id[b] for b in adata.obs[batch_key].cat.categories if b != batch_ref
        ]

        return x, covariates, batches, other_batches, batch_to_id

    return x, covariates, torch.empty((adata.n_obs,)), [], {}
